import 'package:localstorage/localstorage.dart';
import 'package:oc_front/core/conf/conf_reader.dart';
import 'package:oc_front/core/services/api_service.dart';
import 'package:oc_front/core/services/perms_service.dart';
import 'package:oc_front/main.dart';
import 'package:oc_front/models/response.dart';

class AuthService {
  static var conf = AppConfig();
  static var isAuth = const bool.fromEnvironment('AUTH_MODE', defaultValue: true);
  static const _clientID = String.fromEnvironment('CLIENT_ID', defaultValue: 'test-client');
  static APIService<SimpleData>? service;

  static Future<void> init() async {
    service ??= APIService<SimpleData>(baseURL: 
      const String.fromEnvironment("HOST", defaultValue: "http://localhost:8000") + conf.get('AUTH_HOST', defaultValue: '/auth'));
    if (!isAuth) {
      return;
    }
    PermsService.init(localStorage.getItem('accessToken') ?? "");
    bool ok = await introspect().catchError((e) => false);
    if (ok) {
      var now = DateTime.now();
      var expires = DateTime.parse(localStorage.getItem('expiresIn') ??
          DateTime.now().toIso8601String());
      var duration = expires.difference(now);
      refresh(localStorage.getItem('accessToken') ?? "",
          localStorage.getItem('username') ?? "", duration);
    } else {
      localStorage.setItem('accessToken', '');
      localStorage.setItem('username', '');
      localStorage.setItem('expiresIn', '');
    }
  }

  static bool isConnected() {
    if (!isAuth) {
      return true;
    }
    return (localStorage.getItem('accessToken') ?? "") != "" 
    && localStorage.getItem('username') != "";
  }

  static String? getUsername() {
    if (!isAuth) {
      return "no auth user";
    }
    return localStorage.getItem('username') ?? "unknown";
  }

  static Future<void> login(String username, String password) async {
    var token = await service!.post("/login?client_id=$_clientID", <String, dynamic> {
      "username": username,
      "password": password
    }, null);
    if (token.code == 200 && token.data != null) {
      localStorage.setItem('accessToken', token.data?.value['access_token']);
      localStorage.setItem('tokenType', token.data?.value['token_type']);
      localStorage.setItem('username', username);
      localStorage.setItem('expiresIn', DateTime.now().add(
        Duration(seconds: token.data?.value['expires_in'])).toIso8601String());
      mainKey?.currentState?.setState(() {});
      PermsService.init(token.data?.value['access_token']);
      refresh(token.data?.value['access_token'] ?? "", username,
          Duration(seconds: token.data?.value['expires_in']));
    }
  }

  static Future<void> logout() async {
    var token = await service!.delete("/logout?client_id=$_clientID", null);
    if (token.code == 200) {
      localStorage.setItem('accessToken', '');
      localStorage.setItem('username', '');
      localStorage.setItem('expiresIn', '');
      PermsService.clear();
    }
  }

  static Future<bool> introspect() async {
    if (!isConnected()) {
      return false;
    }
    // ignore: invalid_return_type_for_catch_error
    var isIntrospected = await service!.get("/introspect", true, null).catchError((e) => mainKey?.currentState?.setState(() {}));
    return isIntrospected.code == 200;
  }

  static Future<void> refresh(
      String accessToken, String username, Duration duration) async {
    Future.delayed(duration, () {
      service!.post("/refresh?client_id=$_clientID", <String, dynamic> {
        "access_token": accessToken,
        "username": username
      }, null).then((token) {
        if (token.code == 200 && token.data != null) {
          PermsService.init(token.data?.value['access_token']);
          localStorage.setItem(
              'accessToken', token.data?.value['access_token']);
          localStorage.setItem('username', username);
          localStorage.setItem(
              'expiresIn',
              DateTime.now()
                  .add(Duration(seconds: token.data?.value['expires_in']) -
                      Duration(seconds: 10))
                  .toIso8601String());
        }
      });
    });
  }
}
