import 'dart:convert';

import 'dart:ui' as ui;
import 'dart:math' as math;
import 'package:flutter/material.dart';
import 'package:flutter_flow_chart/flutter_flow_chart.dart';
import 'package:flutter_flow_chart/src/ui/segment_handler.dart';


enum ArrowDash {
  /// Arrow pointing from the source to the destination.
  line,
  largeDash,
  mediumDash,
  smallDash,
  heavyDotted,
  mediumDotted,
  lightDotted,
}

double spaceArrowDash(ArrowDash dash) {
  if (dash == ArrowDash.line) { return 0; 
  } else if (dash == ArrowDash.largeDash) { return 12; 
  } else if (dash == ArrowDash.mediumDash) { return 10; 
  } else if (dash == ArrowDash.smallDash) { return 5; 
  } else if (dash == ArrowDash.heavyDotted) { return 2.5; 
  } else if (dash == ArrowDash.mediumDotted) { return 5; 
  } else if (dash == ArrowDash.lightDotted) { return 10; }
  return 0;
}

double widthArrowDash(ArrowDash dash) {
  if (dash == ArrowDash.line) { return 0; 
  } else if (dash == ArrowDash.largeDash) { return 7; 
  } else if (dash == ArrowDash.mediumDash) { return 5; 
  } else if (dash == ArrowDash.smallDash) { return 2.5; 
  } else if (dash == ArrowDash.heavyDotted) { return 1; 
  } else if (dash == ArrowDash.mediumDotted) { return 1; 
  } else if (dash == ArrowDash.lightDotted) { return 1; }
  return 0;
}


enum ArrowDirection {
  /// Arrow pointing from the source to the destination.
  forward,

  /// Arrow pointing from the destination to the source.
  backward,
  bidirectionnal,
}
/// Arrow style enumeration
enum ArrowStyle {
  /// A curved arrow which points nicely to each handlers
  curve,

  /// A segmented line where pivot points can be added and curvature between
  /// them can be adjusted with a tension.
  segmented,

  /// A rectangular shaped line.
  rectangular,
}

/// Arrow parameters used by [DrawArrow] widget
class ArrowParams extends ChangeNotifier {
  ///
  ArrowParams({
    this.dashWidth = 0,
    this.dashSpace = 0,
    this.backwardWidth = 10,
    this.forwardWidth = 10,
    this.thickness = 1.7,
    this.headRadius = 6,
    double tailLength = 25.0,
    this.color = Colors.black,
    this.style,
    this.tension = 1.0,
    this.direction = ArrowDirection.forward,
    this.startArrowPosition = Alignment.centerRight,
    this.endArrowPosition = Alignment.centerLeft,
  }) : _tailLength = tailLength;
    ///
  Map<String, dynamic> toMap() {
    return <String, dynamic>{
      'thickness': thickness,
      'head_radius': headRadius,
      'tail_length': _tailLength,
      'color': color.value,
      'arrow_style': style?.index,
      'tension': tension,
      'direction': direction?.index,
      'dash_width': dashWidth,
      'dash_space': dashSpace,
      'backward_arrow_width': backwardWidth,
      'forward_arrow_width': forwardWidth,
      'start_arrow_position_x': startArrowPosition.x,
      'start_arrow_position_y': startArrowPosition.y,
      'end_arrow_position_x': endArrowPosition.x,
      'end_arrow_position_y': endArrowPosition.y,
    };
  }
  ///
  factory ArrowParams.fromMap(Map<String, dynamic> map) {
    return ArrowParams(
      dashSpace: map['dash_space'] as double,
      dashWidth: map['dash_width'] as double,
      backwardWidth: map['backward_arrow_width'] as double,
      forwardWidth: map['forward_arrow_width'] as double,
      thickness: map['thickness'] as double,
      direction: ArrowDirection.values[map['direction'] as int? ?? 0],
      headRadius: map['head_radius'] as double? ?? 6.0,
      tailLength: map['tail_length'] as double? ?? 25.0,
      color: Color(map['color'] as int),
      style: ArrowStyle.values[map['arrow_style'] as int? ?? 0],
      tension: map['tension'] as double? ?? 1,
      startArrowPosition: Alignment(
        map['start_arrow_position_x'] as double,
        map['start_arrow_position_y'] as double,
      ),
      endArrowPosition: Alignment(
        map['end_arrow_position_x'] as double,
        map['end_arrow_position_y'] as double,
      ),
    );
  }

  ///
  factory ArrowParams.fromJson(String source) =>
      ArrowParams.fromMap(json.decode(source) as Map<String, dynamic>);
  /// Arrow thickness.
  double thickness;
  double forwardWidth;
  double backwardWidth; 
  /// The radius of arrow tip.
  double headRadius;

  /// Arrow color.
  Color color;

  /// The start position alignment.
  final Alignment startArrowPosition;

  /// The end position alignment.
  final Alignment endArrowPosition;

  /// The tail length of the arrow.
  double _tailLength;

  /// The style of the arrow.
  ArrowStyle? style;

  ArrowDirection? direction;
  double dashWidth = 0;
  double dashSpace = 0;
  /// The curve tension for pivot points when using [ArrowStyle.segmented].
  /// 0 means no curve on segments.
  double tension;

  ///
  ArrowParams copyWith({
    double? thickness,
    Color? color,
    ArrowStyle? style,
    ArrowDirection? direction,
    double? tension,
    double? dashWidth,
    double? dashSpace,
    Alignment? startArrowPosition,
    Alignment? endArrowPosition,
    double? backwardWidth,
    double? forwardWidth,
  }) {
    return ArrowParams(
      backwardWidth: backwardWidth ?? this.backwardWidth,
      forwardWidth: forwardWidth ?? this.forwardWidth,
      thickness: thickness ?? this.thickness,
      color: color ?? this.color,
      style: style ?? this.style,
      dashSpace: dashSpace ?? this.dashSpace,
      dashWidth: dashWidth ?? this.dashWidth,
      direction: direction ?? this.direction,
      tension: tension ?? this.tension,
      startArrowPosition: startArrowPosition ?? this.startArrowPosition,
      endArrowPosition: endArrowPosition ?? this.endArrowPosition,
    );
  }
  ///
  String toJson() => json.encode(toMap());

  ///
  void setScale(double currentZoom, double factor) {
    thickness = thickness / currentZoom * factor;
    headRadius = headRadius / currentZoom * factor;
    _tailLength = _tailLength / currentZoom * factor;
    notifyListeners();
  }

  ///
  double get tailLength => _tailLength;
}

/// Notifier to update arrows position, starting/ending points and params
class DrawingArrow extends ChangeNotifier {
  DrawingArrow._();

  /// Singleton instance of this.
  static final instance = DrawingArrow._();

  /// Arrow parameters.
  ArrowParams params = ArrowParams();
  List<dynamic> env = [];
  List<dynamic> infos = [];
  String fromID = "";
  String toID = "";

  /// Sets the parameters.
  void setParams(ArrowParams params) {
    this.params = params;
    notifyListeners();
  }

  /// Starting arrow offset.
  Offset from = Offset.zero;

  ///
  void setFrom(Offset from) {
    this.from = from;
    
    notifyListeners();
  }

  /// Ending arrow offset.
  Offset to = Offset.zero;

  ///
  void setTo(Offset to) {
    this.to = to;
    notifyListeners();
  }

  ///
  bool isZero() {
    return from == Offset.zero && to == Offset.zero;
  }

  ///
  void reset() {
    params = ArrowParams();
    from = Offset.zero;
    to = Offset.zero;
    notifyListeners();
  }
}

/// Draw arrow from [srcElement] to [destElement]
/// using [arrowParams] parameters
// ignore: must_be_immutable
class DrawArrow extends StatefulWidget {
  ChartWidgetState flow;
  ///
  DrawArrow({
    required this.flow,
    required this.index,
    required this.srcElement,
    required this.destElement,
    required List<Pivot> pivots,
    this.infos = const [],
    this.env = const [],
    super.key,
    ArrowParams? arrowParams,
  })  : arrowParams = arrowParams ?? ArrowParams(),
        pivots = PivotsNotifier(pivots);

  final int index;
   List<dynamic> env = [];
  List<dynamic> infos = [];
  ///
  final ArrowParams arrowParams;

  ///
  final FlowElement srcElement;

  ///
  final FlowElement destElement;

  ///
  final PivotsNotifier pivots;

  @override
  State<DrawArrow> createState() => DrawArrowState();
}

class DrawArrowState extends State<DrawArrow> {
  @override
  void initState() {
    super.initState();
    widget.srcElement.addListener(_elementChanged);
    widget.destElement.addListener(_elementChanged);
    widget.pivots.addListener(_elementChanged);
  }

  @override
  void dispose() {
    widget.srcElement.removeListener(_elementChanged);
    widget.destElement.removeListener(_elementChanged);
    widget.pivots.removeListener(_elementChanged);
    super.dispose();
  }

  void _elementChanged() {
    if (mounted) setState(() {});
  }

  @override
  Widget build(BuildContext context) {
    var from = Offset.zero;
    var to = Offset.zero;
    from = Offset(
      widget.srcElement.position.dx +
          widget.srcElement.handlerSize / 2.0 +
          (widget.srcElement.size.width *
              ((widget.arrowParams.startArrowPosition.x + 1) / 2)),
      widget.srcElement.position.dy +
          widget.srcElement.handlerSize / 2.0 +
          (widget.srcElement.size.height *
              ((widget.arrowParams.startArrowPosition.y + 1) / 2)),
    );
    to = Offset(
      widget.destElement.position.dx +
          widget.destElement.handlerSize / 2.0 +
          (widget.destElement.size.width *
              ((widget.arrowParams.endArrowPosition.x + 1) / 2)),
      widget.destElement.position.dy +
          widget.destElement.handlerSize / 2.0 +
          (widget.destElement.size.height *
              ((widget.arrowParams.endArrowPosition.y + 1) / 2)),
    );
    GlobalKey<GraphParamsWidgetState> key = GlobalKey<GraphParamsWidgetState>();
    return Stack( children : [
        GraphParamsWidget(key: key, element: widget.srcElement, index: widget.index, comp: widget.flow),
        RepaintBoundary(
        child: Builder(
          builder: (context) {
            var painter = ArrowPainter(
                connIndex: widget.index,
                infos: widget.infos,
                env: widget.env,
                elementIndex: widget.flow.widget.dashboard.elements.indexOf(widget.srcElement),
                fromID: "${widget.srcElement.id}_${widget.index}",
                toID: "${widget.destElement.id}_${widget.index}",
                isSelected: widget.flow.widget.dashboard.arrowsSelected.where(
                  (element) => element.fromID == "${widget.srcElement.id}_${widget.index}").isNotEmpty,
                params: widget.arrowParams,
                from: from, to: to,
                pivots: widget.pivots.value,
              );
            if ( widget.flow.widget.dashboard.arrows.where(
                  (element) => element.fromID == "${widget.srcElement.id}_${widget.index}").isEmpty) {
                widget.flow.widget.dashboard.addArrows(painter, context); 
                widget.flow.widget.dashboard.saveDash(widget.flow.widget.dashboard.id, context);
            } else {
              var i = widget.flow.widget.dashboard.arrows.indexWhere(
                  (element) => element.fromID == "${widget.srcElement.id}_${widget.index}");
              painter.isSelected = widget.flow.widget.dashboard.arrows[i].isSelected;
              widget.flow.widget.dashboard.arrows[i] = painter;
            }
            return CustomPaint(
              painter: painter,
              child: Container(),
            );
          },
        ),
      )
    ]);
  }
}

// ignore: must_be_immutable
class GraphParamsWidget extends StatefulWidget {
  ChartWidgetState comp;
  bool isShowed = false;
  Offset? position;
  FlowElement element;
  int index;
  GraphParamsWidget({ super.key, required this.element, required this.comp,
  required this.index });
  @override GraphParamsWidgetState createState() => GraphParamsWidgetState();
}
class GraphParamsWidgetState extends State<GraphParamsWidget> {  
  @override Widget build(BuildContext context) {
    return !widget.isShowed ? Container() : Positioned(
    top: (widget.position?.dy ?? 0) - 5, left: (widget.position?.dx ?? 0) - 5, 
    child: MouseRegion( cursor: SystemMouseCursors.click,
      onHover: (event) => setState(() { widget.isShowed = true; }),
      child: Row(children: [
        IconButton(onPressed: () {
          widget.comp.setState(() {
            widget.comp.widget.dashboard.removeArrows((el) => el.fromID == "${widget.element.id}${widget.index}", context);
            widget.element.next.removeAt(widget.index);
          });
        }, icon: const Icon(Icons.delete))
      ],)));
  }
}

// ignore: must_be_immutable
class ArrowInfoWidget extends StatefulWidget {
  Dashboard dashboard;
  ArrowInfoWidget ({ super.key, required this.dashboard });
  @override ArrowInfoWidgetState createState() => ArrowInfoWidgetState();
}
class ArrowInfoWidgetState extends State<ArrowInfoWidget> {
  @override Widget build(BuildContext context) {
    return SingleChildScrollView( child: Column(children: [
      Container(height: 50, 
       decoration: BoxDecoration(color: widget.dashboard.midDashColor, border: const Border(bottom: BorderSide(color: Colors.grey, width: 1))),
       child: const Center( child: Text("<Arrow> Style", style: TextStyle(fontSize: 20)))),
      Container(height: 50, 
       decoration: BoxDecoration(color: widget.dashboard.midDashColor, border: const Border(bottom: BorderSide(color: Colors.grey, width: 1))),
       child: const Row(children: [],)
      ),
    ],) );
  }
}

/// Paint the arrow connection taking in count the
/// [ArrowParams.startArrowPosition] and
/// [ArrowParams.endArrowPosition] alignment.
class ArrowPainter extends CustomPainter {
  ArrowPainter({
    this.elementIndex,
    this.connIndex,
    this.toID = "",
    this.fromID = "",
    this.infos = const [],
    this.env  = const [],
    this.isSelected = false,
    required this.params,
    this.from = Offset.zero,
    this.to = Offset.zero,
    List<Pivot>? pivots, 
  }) : pivots = pivots ?? [];

  List<dynamic> env;
  List<dynamic> infos;
  ArrowParams params;
  String toID;
  String fromID;
  Offset to;
  Offset from;
  int? connIndex;
  int? elementIndex;
  bool isSelected = false;
  ///
  Path path = Path();
  Path dashed = Path();
  Path subPath = Path();

  ///
  final List<List<Offset>> lines = [];

  ///
  final List<Pivot> pivots;

  late DashedPathProperties _dashedPathProperties;

  final arrowSize = 15;
  final arrowAngle=  25 * math.pi / 180;

  ArrowPainter deserialize(Map<String, dynamic> map) {
    infos = map['infos'];
    env = map['env'];
    params = ArrowParams.fromMap(map['params']);
    fromID = map['from']['id'];
    toID = map['to']['id'];
    from = Offset(map['from']['x'], map['from']['y']);
    to = Offset(map['to']['x'], map['to']['y']);
    return this;
  }
  Map<String, dynamic> serialize() {
    Map<String, dynamic> graphElement = {};
    graphElement['infos'] = infos;
    graphElement['env'] = env;
    graphElement['from'] = { "id" : fromID.split("_")[0], "x" : from.dx, "y" : from.dy };
    graphElement['to'] = { "id" : toID.split("_")[0], "x" : to.dx, "y" : to.dy };
    graphElement['params'] = params.toMap();
    return graphElement;
  }

  factory ArrowPainter.fromMap(Map<String, dynamic> map) {
    final e = ArrowPainter(
      connIndex: map['connIndex'] as int,
      elementIndex: map['elementIndex'] as int,
      toID: map['toID'] != null ? map['toID'] as String : "",
      fromID: map['fromID'] as String,
      infos: map['infos'] as List,
      env: map['env'] as List,
      isSelected: map['isSelected'] != null ? map['isSelected'] as bool : false,
      params: ArrowParams.fromMap(map['params']),
      from: Offset(
        map['fromDx'] as double,
        map['fromDy'] as double,
      ),
      to: Offset(
        map['toDx'] as double,
        map['toDy'] as double,
      ),
      pivots: (map['pivots'] as List).map<Pivot>((e) => Pivot.fromMap(e as Map<String, dynamic>)).toList(),
    );
    return e;
  }

  Map<String, dynamic> toMap() {
    return {
        'toID': toID,
        'fromID': fromID,
        'elementIndex': elementIndex,
        'connIndex': connIndex,
        'isSelected': isSelected,
        'params': params.toMap(),
        "fromDx" : from.dx,
        "fromDy" : from.dy,
        "toDx" : to.dx,
        "toDy" : to.dy,
        "infos" : infos,
        "env" : env,
        'pivots': pivots.map((e) => e.toMap()).toList(),
      };
  }

  

  @override
  void paint(Canvas canvas, Size size) {
    if (params.dashSpace > 0) {
      _dashedPathProperties = DashedPathProperties(
        path: Path(),
        dashLength: params.dashWidth,
        dashGapLength: params.dashSpace,
      );
    }
    final paint = Paint()..strokeWidth = params.thickness;
    from = Offset(from.dx - params.headRadius - 2, from.dy - params.headRadius - 2 );
    to = Offset(to.dx - params.headRadius - 3, to.dy - params.headRadius - 2 );

    if (params.style == ArrowStyle.curve) { drawCurve(canvas, paint);
    } else if (params.style == ArrowStyle.segmented) { drawLine();
    } else if (params.style == ArrowStyle.rectangular) { drawRectangularLine(canvas, paint); }

    paint
      ..color = isSelected ? Colors.red : params.color
      ..style = PaintingStyle.stroke;
    canvas.drawPath(dashed, paint);
    canvas.drawPath(subPath, paint);
  }

  Path drawArrow(Offset a, Offset b, Alignment? start, bool isCurve, bool backward) {
    double arrowSize = 0;
    if (backward) { arrowSize = params.backwardWidth; 
    } else { arrowSize = params.forwardWidth; }
    const arrowAngle = math.pi / 6;
    final dX = b.dx - a.dx;
    final dY = b.dy - a.dy;
    double angle = math.atan2(dY, dX);

    if (start == Alignment.bottomCenter) { angle = -33 - (isCurve ? 3 : 0); 
    } else if (start == Alignment.topCenter) { angle = 33 + (isCurve ? 3 : 0); 
    } else if (start == Alignment.centerRight) { angle = 66; 
    } else if (start == Alignment.centerLeft) { angle = 0; }

    subPath.moveTo(b.dx - arrowSize * math.cos(angle - arrowAngle),
        b.dy - arrowSize * math.sin(angle - arrowAngle));
    subPath.lineTo(b.dx, b.dy);
    subPath.lineTo(b.dx - arrowSize * math.cos(angle + arrowAngle),
        b.dy - arrowSize * math.sin(angle + arrowAngle));
    return subPath;
  }

  /// Draw a segmented line with a tension between points.
  void drawLine() {
    final points = [from];
    for (final pivot in pivots) {
      points.add(pivot.pivot);
    }
    points.add(to);

    path.moveTo(points.first.dx, points.first.dy);

    for (var i = 0; i < points.length - 1; i++) {
      final p0 = (i > 0) ? points[i - 1] : points[0];
      final p1 = points[i];
      final p2 = points[i + 1];
      final p3 = (i != points.length - 2) ? points[i + 2] : p2;

      final cp1x = p1.dx + (p2.dx - p0.dx) / 6 * params.tension;
      final cp1y = p1.dy + (p2.dy - p0.dy) / 6 * params.tension;

      final cp2x = p2.dx - (p3.dx - p1.dx) / 6 * params.tension;
      final cp2y = p2.dy - (p3.dy - p1.dy) / 6 * params.tension;

      path.cubicTo(cp1x, cp1y, cp2x, cp2y, p2.dx, p2.dy);
    }

    if (params.dashSpace > 0) {
      dashed = _getDashedPath(path, params.dashWidth, params.dashSpace);
    } else { dashed = path; }
    if (ArrowDirection.forward  == params.direction || ArrowDirection.bidirectionnal == params.direction) { 
      subPath = drawArrow(from, to, null, false, false); 
    }
    if (ArrowDirection.backward  == params.direction || ArrowDirection.bidirectionnal == params.direction) {  
      subPath = drawArrow(to, from, null, false, true); }
  }

  /// Draw a rectangular line
  void drawRectangularLine(Canvas canvas, Paint paint) {
    // calculating offsetted pivot
    var pivot1 = Offset(from.dx, from.dy);
    if (params.startArrowPosition.y == 1) {
      pivot1 = Offset(from.dx, from.dy + params.tailLength);
    } else if (params.startArrowPosition.y == -1) {
      pivot1 = Offset(from.dx, from.dy - params.tailLength);
    }

    final pivot2 = Offset(to.dx, pivot1.dy);

    path
      ..moveTo(from.dx, from.dy)
      ..lineTo(pivot1.dx, pivot1.dy)
      ..lineTo(pivot2.dx, pivot2.dy)
      ..lineTo(to.dx, to.dy);

    lines.addAll([
      [from, pivot2],
      [pivot2, to],
    ]);
    if (params.dashSpace != 0) {
      dashed = _getDashedPath(path, params.dashWidth, params.dashSpace);
    } else { dashed = path; }
    Alignment start = Alignment.centerLeft;
    if (pivot2.dy != to.dy) {
      if (to.dy < pivot2.dy) { start = Alignment.bottomCenter;
      } else { start = Alignment.topCenter; }
    }
    if (pivot2.dx != to.dx) {
      if (to.dx < pivot2.dx) { start = Alignment.centerRight;
      } else { start = Alignment.centerLeft; }
    }
    if (ArrowDirection.forward  == params.direction || ArrowDirection.bidirectionnal == params.direction) { 
      subPath = drawArrow(from, to, start, false, false); 
    }
    if (ArrowDirection.backward  == params.direction || ArrowDirection.bidirectionnal == params.direction) {  
      subPath = drawArrow(to, from, params.startArrowPosition, false, true); }
  }

  /// Draws a curve starting/ending the handler linearly from the center
  /// of the element.
  void drawCurve(Canvas canvas, Paint paint) {
    var distance = 0.0;
    var dx = 0.0;
    var dy = 0.0;

    final p0 = Offset(from.dx, from.dy);
    final p4 = Offset(to.dx, to.dy);
    distance = (p4 - p0).distance / 3;
    
    // checks for the arrow direction
    if (params.startArrowPosition.x > 0) {
      dx = distance;
    } else if (params.startArrowPosition.x < 0) {
      dx = -distance;
    }
    if (params.startArrowPosition.y > 0) {
      dy = distance;
    } else if (params.startArrowPosition.y < 0) {
      dy = -distance;
    }
    final p1 = Offset(from.dx + dx, from.dy + dy);
    dx = 0;
    dy = 0;

    // checks for the arrow direction
    if (params.endArrowPosition.x > 0) {
      dx = distance;
    } else if (params.endArrowPosition.x < 0) {
      dx = -distance;
    }
    if (params.endArrowPosition.y > 0) {
      dy = distance;
    } else if (params.endArrowPosition.y < 0) {
      dy = -distance;
    }
    final p3 = params.endArrowPosition == Alignment.center
        ? Offset(to.dx, to.dy)
        : Offset(to.dx + dx, to.dy + dy);
    final p2 = Offset(
      p1.dx + (p3.dx - p1.dx) / 2,
      p1.dy + (p3.dy - p1.dy) / 2,
    );

    path
      ..moveTo(p0.dx, p0.dy)
      ..conicTo(p1.dx, p1.dy, p2.dx, p2.dy, 1)
      ..conicTo(p3.dx, p3.dy, p4.dx, p4.dy, 1);
    if (params.dashSpace > 0) {
      dashed = _getDashedPath(path, params.dashWidth, params.dashSpace);
    } else { dashed = path; }
    Alignment start = Alignment.center;
    if (p3.dy != p4.dy) {
      if (p4.dy < p3.dy) { start = Alignment.bottomCenter;
      } else { start = Alignment.topCenter; }
    }
    if (p3.dx != p4.dx) {
      if (p4.dx < p3.dx) { start = Alignment.centerRight;
      } else { start = Alignment.centerLeft; }
    }
    if (ArrowDirection.forward  == params.direction || ArrowDirection.bidirectionnal == params.direction) { 
      subPath = drawArrow(from, to, start, start == Alignment.center, false); 
    }
    if (ArrowDirection.backward  == params.direction || ArrowDirection.bidirectionnal == params.direction) {  
      subPath = drawArrow(to, from, params.startArrowPosition, params.startArrowPosition == Alignment.center, true); 
    }
  }

  @override
  bool shouldRepaint(ArrowPainter oldDelegate) {
    return true;
  }

  Path _getDashedPath(
      Path originalPath, double dashLength, double dashGapLength) {
    final metricsIterator = originalPath.computeMetrics().iterator;
    while (metricsIterator.moveNext()) {
      final metric = metricsIterator.current;
      _dashedPathProperties.extractedPathLength = 0.0;
      while (_dashedPathProperties.extractedPathLength < metric.length) {
        if (_dashedPathProperties.addDashNext) {
          _dashedPathProperties.addDash(metric, dashLength);
        } else {
          _dashedPathProperties.addDashGap(metric, dashGapLength);
        }
      }
    }
    return _dashedPathProperties.path;
  }

  bool isLine(Offset point) {
    final metrics = path.computeMetrics(forceClosed: false);
    for (final metric in metrics) {
      final length = metric.length;
      const steps = 100;
      for(int i = 0; i <= steps; i++) {
        final distance = length * i / steps;
        final tangent = metric.getTangentForOffset(distance);
        if (tangent != null) {
          final position = tangent.position;
          if ((position - point).distance <= 10) {
            return true;
          }
        }
      }
    }
    return false;
  }
/*
  bool isLine(Offset position, String name) {
    position = Offset(position.dx, position.dy + 10);
    print("${position.dx} ${position.dy}");
    for (double i=-5; i < 5; i++) {
      for (double y=-5; y < 5; y++) {
        var pos = position + Offset(i, y);
        if (path.contains(pos)) {
          return true;
        }
      }
    }
    return false;
  }
*/
  @override
  bool? hitTest(Offset position) {
    /* graphkey?.currentState?.widget.isShowed = isLine(position);
    if (isLine(position)) { graphkey?.currentState?.widget.position = graphkey?.currentState?.widget.position ?? position; }
    Future.delayed(Duration(seconds: 1), () {
      if (graphkey != null && graphkey!.currentState != null && !graphkey!.currentState!.widget.isShowed) {
        graphkey?.currentState?.widget.position = null;
      }
    });
    graphkey?.currentState?.setState(() {});*/
    return false;
  }
}

/// Notifier for pivot points.
class PivotsNotifier extends ValueNotifier<List<Pivot>> {
  ///
  PivotsNotifier(super.value) {
    for (final pivot in value) {
      pivot.addListener(notifyListeners);
    }
  }

  /// Add a pivot point.
  void add(Pivot pivot) {
    value.add(pivot);
    pivot.addListener(notifyListeners);
    notifyListeners();
  }

  /// Remove a pivot point.
  void remove(Pivot pivot) {
    value.remove(pivot);
    pivot.removeListener(notifyListeners);
    notifyListeners();
  }

  /// Insert a pivot point.
  void insert(int index, Pivot pivot) {
    value.insert(index, pivot);
    pivot.addListener(notifyListeners);
    notifyListeners();
  }

  /// Remove a pivot point by its index.
  void removeAt(int index) {
    value.removeAt(index).removeListener(notifyListeners);
    notifyListeners();
  }
}

class DashedPathProperties {
  double extractedPathLength;
  Path path;

  final double _dashLength;
  double _remainingDashLength;
  double _remainingDashGapLength;
  bool _previousWasDash = false;

  DashedPathProperties({
    required this.path,
    required double dashLength,
    required double dashGapLength,
  })  : 
        _dashLength = dashLength,
        _remainingDashLength = dashLength,
        _remainingDashGapLength = dashGapLength,
        _previousWasDash = false,
        extractedPathLength = 0.0;

  bool get addDashNext {
    if (_remainingDashLength != _dashLength || !_previousWasDash) {
      return true;
    }
    return false;
  }

  void addDash(ui.PathMetric metric, double dashLength) {
    // Calculate lengths (actual + available)
    final end = _calculateLength(metric, _remainingDashLength);
    final availableEnd = _calculateLength(metric, dashLength);
    // Add path
    final pathSegment = metric.extractPath(extractedPathLength, end);
    path.addPath(pathSegment, Offset.zero);
    // Update
    final delta = _remainingDashLength - (end - extractedPathLength);
    _remainingDashLength = _updateRemainingLength(
      delta: delta,
      end: end,
      availableEnd: availableEnd,
      initialLength: dashLength,
    );
    extractedPathLength = end;
    _previousWasDash = true;
  }

  void addDashGap(ui.PathMetric metric, double dashGapLength) {
    // Calculate lengths (actual + available)
    final end = _calculateLength(metric, _remainingDashGapLength);
    final availableEnd = _calculateLength(metric, dashGapLength);
    // Move path's end point
    ui.Tangent tangent = metric.getTangentForOffset(end)!;
    path.moveTo(tangent.position.dx, tangent.position.dy);
    // Update
    final delta = end - extractedPathLength;
    _remainingDashGapLength = _updateRemainingLength(
      delta: delta,
      end: end,
      availableEnd: availableEnd,
      initialLength: dashGapLength,
    );
    extractedPathLength = end;
    _previousWasDash = false;
  }

  double _calculateLength(ui.PathMetric metric, double addedLength) {
    return math.min(extractedPathLength + addedLength, metric.length);
  }

  double _updateRemainingLength({
    required double delta,
    required double end,
    required double availableEnd,
    required double initialLength,
  }) {
    return (delta > 0 && availableEnd == end) ? delta : initialLength;
  }
}
