import 'package:flutter/material.dart';
import 'package:flutter_flow_chart/flutter_flow_chart.dart';
import 'package:flutter_flow_chart/src/objects/diamond_widget.dart';
import 'package:flutter_flow_chart/src/objects/hexagon_widget.dart';
import 'package:flutter_flow_chart/src/objects/oval_widget.dart';
import 'package:flutter_flow_chart/src/objects/parallelogram_widget.dart';
import 'package:flutter_flow_chart/src/objects/rectangle_widget.dart';
import 'package:flutter_flow_chart/src/objects/storage_widget.dart';
import 'package:flutter_flow_chart/src/ui/draw_arrow.dart';
import 'package:flutter_flow_chart/src/ui/element_handlers.dart';
import 'package:flutter_flow_chart/src/ui/resize_widget.dart';
import 'package:flutter_flow_chart/src/objects/any_widget.dart';

/// Widget that use [element] properties to display it on the dashboard scene
class ElementWidget<T extends FlowData> extends StatefulWidget {
  Widget? bottomLeftBadge;
  Widget? bottomRightBadge;
  Widget? topLeftBadge;
  Widget? topRightBadge;
  ///
  ElementWidget({
    required this.dashboard,
    required this.element,
    super.key,
    this.bottomLeftBadge,
    this.bottomRightBadge,
    this.topLeftBadge,
    this.topRightBadge,
    this.onElementPressed,
    this.onElementSecondaryTapped,
    this.onElementLongPressed,
    this.onElementSecondaryLongTapped,
    this.onHandlerPressed,
    this.onHandlerSecondaryTapped,
    this.onHandlerLongPressed,
    this.onHandlerSecondaryLongTapped,
  });
  Color? dashColor;
  ///
  final Dashboard dashboard;

  ///
  final FlowElement element;
  ///
  bool isHovered = false;

  ///
  final void Function(BuildContext context, Offset position)? onElementPressed;

  ///
  final void Function(BuildContext context, Offset position)?
      onElementSecondaryTapped;

  ///
  final void Function(BuildContext context, Offset position)?
      onElementLongPressed;

  ///
  final void Function(BuildContext context, Offset position)?
      onElementSecondaryLongTapped;

  ///
  final void Function(
    BuildContext context,
    Offset position,
    Handler handler,
    FlowElement element,
  )? onHandlerPressed;

  ///
  final void Function(
    BuildContext context,
    Offset position,
    Handler handler,
    FlowElement element,
  )? onHandlerSecondaryTapped;

  ///
  final void Function(
    BuildContext context,
    Offset position,
    Handler handler,
    FlowElement element,
  )? onHandlerLongPressed;

  ///
  final void Function(
    BuildContext context,
    Offset position,
    Handler handler,
    FlowElement element,
  )? onHandlerSecondaryLongTapped;

  @override
  State<ElementWidget> createState() => ElementWidgetState<T>();
}

class ElementWidgetState<T extends FlowData>  extends State<ElementWidget> {
  // local widget touch position when start dragging
  Offset delta = Offset.zero;

  @override
  void initState() {
    super.initState();
    widget.element.addListener(_elementChanged);
  }

  @override
  void dispose() {
    widget.element.removeListener(_elementChanged);
    super.dispose();
  }

  void _elementChanged() {
    setState(() {});
  }

  @override
  Widget build(BuildContext context) {
    Widget element;

    switch (widget.element.kind) {
      case ElementKind.diamond:
        element = DiamondWidget(element: widget.element);
      case ElementKind.storage:
        element = StorageWidget(element: widget.element);
      case ElementKind.oval:
        element = OvalWidget(element: widget.element);
      case ElementKind.parallelogram:
        element = ParallelogramWidget(element: widget.element);
      case ElementKind.hexagon:
        element = HexagonWidget(element: widget.element);
      case ElementKind.rectangle:
        element = RectangleWidget(element: widget.element);
      case ElementKind.widget:
        if (widget.element.widget == null) { element = RectangleWidget(element: widget.element);
        } else { element = AnyWidget(element: widget.element); }
    }
    var tapLocation = Offset.zero;
    var secondaryTapDownPos = Offset.zero;
    Widget w = InkWell(
        onTapDown: (details) => tapLocation = details.globalPosition,
        onSecondaryTapDown: (details) =>
            secondaryTapDownPos = details.globalPosition,
        onTap: () {
          setState(() { 
            if (!isCtrl) {
              for (var element in widget.dashboard.elements) { 
                element.isSelected = false; 
                element.dashboard.chartKey.currentState?. setState(() { });
                widget.dashboard.selectedMenuKey.currentState?. setState(() { });
              }
            }
            DrawingArrow.instance.notifyListeners(); 
            widget.isHovered = false;
            widget.element.isSelected = !widget.element.isSelected; 
            for (var sel in widget.dashboard.arrows) { sel.isSelected = false; }
            widget.dashboard.selectedMenuKey.currentState?. setState(() { });
          });
        },
        onSecondaryTap: () {
          widget.onElementSecondaryTapped?.call(context, secondaryTapDownPos);
        },
        onLongPress: () {
          widget.onElementLongPressed?.call(context, tapLocation);
        },
        /*onSecondaryLongPress: () {
          widget.onElementSecondaryLongTapped?.call(context, secondaryTapDownPos);
        },*/
        child: Listener(
          onPointerDown: (event) {
            delta = event.localPosition;
          },
          child: Draggable<FlowElement>(
            data: widget.element,
            dragAnchorStrategy: childDragAnchorStrategy,
            childWhenDragging: const SizedBox.shrink(),
            feedback: Material(
              color: Colors.transparent,
              child: Padding( padding: EdgeInsets.all(6),
                child: Container(
                  decoration: BoxDecoration(border: Border.all( color: Colors.red, width: 2)),
                  width: widget.element.size.width - 12, 
                  height: widget.element.size.height - 12, child: element)
              ),
            ),
            child: ElementHandlers(
                  dashboard: widget.dashboard,
                  element: widget.element,
                  handlerSize: widget.element.handlerSize,
                  onHandlerPressed: widget.onHandlerPressed,
                  onHandlerSecondaryTapped: widget.onHandlerSecondaryTapped,
                  onHandlerLongPressed: widget.onHandlerLongPressed,
                  onHandlerSecondaryLongTapped: widget.onHandlerSecondaryLongTapped,
                  child: Container( 
                      margin: const EdgeInsets.all(10), // why some change
                      decoration: BoxDecoration(
                        border: Border.all(color: widget.element.isSelected ? Colors.red : widget.dashboard.midDashColor, 
                          width: widget.element.isSelected ? 2 : 1),
                      ),
                      child: InkWell( mouseCursor: SystemMouseCursors.grab, child: element )
                  )
              ),
            onDragStarted: () { 
              widget.element.isSelected = true;
              widget.dashboard.selectedMenuKey.currentState?. setState(() { });
            },
            onDragUpdate: (details) {
              var diff = Offset(widget.element.position.dx, widget.element.position.dy);
              widget.element.isSelected = true;
              widget.element.changePosition(
                details.globalPosition - delta - Offset(
                  widget.dashboard.widthOffset, 50 + widget.dashboard.heightOffset), // todo change
              ); 
              diff = widget.element.position - diff;
              if (widget.element.isSelected) {
                for (var sel in widget.dashboard.elementSelected) {
                  if (widget.element == sel) { continue; }
                  sel.changePosition(sel.position + diff);
                }
              }
            },
            onDragEnd: (details) {
              var diff = Offset(widget.element.position.dx, widget.element.position.dy);
              widget.element.changePosition(details.offset - Offset(
                widget.dashboard.widthOffset, 50 + widget.dashboard.heightOffset
              )); // todo change
              diff = widget.element.position - diff;
              if (widget.element.isSelected) {
                for (var sel in widget.dashboard.elementSelected) {
                  if (widget.element == sel) { continue; }
                  sel.changePosition(sel.position + diff);
                }
              }
              widget.dashboard.saveDash(widget.dashboard.id, context);
            },
          ),
        ),
    );
    w = ResizeWidget(
      comp: this,
      bottomLeftBadge: widget.bottomLeftBadge,
      bottomRightBadge: widget.bottomRightBadge,
      topLeftBadge: widget.topLeftBadge,
      topRightBadge: widget.topRightBadge,
      element: widget.element,
      dashboard: widget.dashboard,
      handlerColor: widget.isHovered ? widget.dashboard.dashColor : Colors.transparent,
      child: w 
    );
    return Transform.translate(
      offset: widget.element.position,
      transformHitTests: true,
      child: MouseRegion( 
        onEnter: (event) { setState(() {  widget.isHovered = true;  }); }, 
        onExit: (event) { setState(() {  widget.isHovered = false; }); },
        child: Wrap( direction: Axis.vertical, children: [ 
          w,         
          Container( 
            constraints: BoxConstraints( minWidth: widget.element.size.width ),
            child: Row( mainAxisAlignment: MainAxisAlignment.center, 
              children: (!widget.isHovered ? [] : [
              IconButton(tooltip: "remove element", onPressed: () {
                widget.dashboard.removeElement(widget.element, context);
              }, icon: Icon(Icons.delete_outline)),
              IconButton(tooltip: "copy element", onPressed: () {
                FlowElement<T> newElement = FlowElement<T>(
                  element: widget.element.element as T?,
                  dashboard: widget.dashboard,
                  kind: widget.element.kind,
                  position: widget.element.position + Offset(widget.element.size.width, widget.element.size.height),
                  size: widget.element.size,
                  widget: widget.element.widget,
                );
                widget.dashboard.addElement(newElement, context);
              }, icon: Icon(Icons.copy, size: 20)),
            ]))
          ),
        ])
      )
    );
  }
}
